@@ -23,6 +23,7 @@ def parse(sql_text, start='sql_script', strict=False):
 
     if strict:
         error_listener = CustomErrorListener()
+        parser.removeErrorListeners()
         parser.addErrorListener(error_listener)
 
     return ast.visit(getattr(parser, start)())
@@ -35,7 +36,7 @@ def parse_from_yaml(fname):
         out[start] = [parse(cmd, start) for cmd in cmds]
     return out
 
-from antlr_ast import AstNode
+from antlr_ast import AstNode, Speaker
 
 class Unshaped(AstNode):
     _fields = ['arr']
@@ -74,10 +75,25 @@ class SelectStmt(AstNode):
         for k in unlist_clauses:
             attr = getattr(query, k, [])
             if isinstance(attr, list):
-                setattr(query, k, attr[0] if len(attr) == 1 else None)
+                setattr(query, k, attr[0] if len(attr) else None)
 
         return query
 
+class Union(AstNode):
+    _fields = ['left', 'op', 'right', 'order_by_clause']        
+
+    @classmethod
+    def _from_subquery_compound(cls, visitor, ctx):
+        # hoists up ORDER BY clauses from the right SELECT statement
+        # since the final ORDER BY applies to the entire statement (not just subquery)
+        union = cls._from_fields(visitor, ctx)
+        if not isinstance(ctx.right, plsqlParser.SubqueryParenContext):
+            order_by = getattr(union.right, 'order_by_clause', None)
+            union.order_by_clause = order_by
+            # remove from right SELECT
+            if order_by: union.right.order_by_clause = None
+
+        return union
 
 class Identifier(AstNode):
     _fields = ['fields']
@@ -103,6 +119,13 @@ class BinaryExpr(AstNode):
 
         return bin_expr
 
+    @classmethod
+    def _from_in_expr(cls, visitor, ctx):
+        bin_expr = cls._from_mod(visitor, ctx)
+        right = visitor.visit(ctx.subquery() or ctx.expression_list())
+        bin_expr.right = right if isinstance(right, list) else right
+        return bin_expr
+
 class UnaryExpr(AstNode):
     _fields = ['op', 'unary_expression->expr']
 
@@ -114,6 +137,20 @@ class SortBy(AstNode):
     _fields = ['expression->expr', 'direction', 'nulls']
     #_rules = ['order_by_elements']
 
+class JoinExpr(AstNode):
+    _fields = ['left', 'join_type', 'table_ref->right',
+               'join_on_part->cond', 
+               # fields below are Oracle specific
+               'join_using_part->using', 'query_partition_clause']
+
+    @classmethod
+    def _from_table_ref(cls, visitor, ctx):
+        join_expr = cls._from_fields(visitor, ctx.join_clause())
+        join_expr._ctx = ctx
+        join_expr.left = visitor.visit(ctx.table_ref())
+
+        return join_expr
+
 from collections.abc import Sequence
 class Call(AstNode):
     _fields = ['name', 'pref', 'args', 'function_argument_analytic->args', 'concatenation->args', 'over_clause']
@@ -174,9 +211,7 @@ class AstVisitor(plsqlVisitor):
         return self.visit(ctx.subquery())
 
     def visitSubqueryCompound(self, ctx):
-        # TODO: here we form UNION statements etc into binary expr, but should
-        #       use a compound statement as in official ast
-        return BinaryExpr._from_fields(self, ctx)
+        return Union._from_subquery_compound(self, ctx)
 
     def visitQuery_block(self, ctx):
         return SelectStmt._from_query_block(self, ctx)
@@ -213,6 +248,9 @@ class AstVisitor(plsqlVisitor):
     def visitModExpr(self, ctx):
         return BinaryExpr._from_mod(self, ctx)
 
+    def visitJoinExpr(self, ctx):
+        return JoinExpr._from_table_ref(self, ctx)
+
     # function calls -------
     def visitAggregate_windowed_function(self, ctx):
         return Call._from_aggregate_call(self, ctx)
@@ -230,10 +268,12 @@ class AstVisitor(plsqlVisitor):
 
     # expression conds
     # TODO replace with for loop over AST classes
+    def visitInExpr(self, ctx):
+        return BinaryExpr._from_in_expr(self, ctx)
+
     visitIsExpr =     visitBinaryExpr
-    visitInExpr =    visitBinaryExpr
-    visitBetweenExpr = visitBinaryExpr
-    visitLikeExpr = visitBinaryExpr
+    visitBetweenExpr = visitModExpr
+    visitLikeExpr = visitModExpr
     visitRelExpr =    visitBinaryExpr
     visitMemberExpr = visitBinaryExpr
     visitCursorExpr = visitUnaryExpr
@@ -249,7 +289,7 @@ class AstVisitor(plsqlVisitor):
         return self.visitChildren(ctx, predicate = lambda n: n is not ctx.WHERE())
 
     def visitFrom_clause(self, ctx):
-        return  self.visitChildren(ctx, predicate = lambda n: n is not ctx.FROM())
+        return  self.visitChildren(ctx, predicate = lambda n: not isinstance(n, Tree.TerminalNode))
 
     def visitLimit_clause(self, ctx):
         return  self.visitChildren(ctx, predicate = lambda n: n is not ctx.LIMIT())
@@ -267,8 +307,16 @@ class AstVisitor(plsqlVisitor):
         return  self.visitChildren(ctx, predicate = lambda n: n is not ctx.HAVING())
 
     def visitExpression_list(self, ctx):
+        return [self.visit(expr) for expr in ctx.expression()]
+        #return self.visitChildren(ctx, predicate = lambda n: not isinstance(n, Tree.TerminalNode))
+
+    def visitJoin_on_part(self, ctx):
         return self.visitChildren(ctx, predicate = lambda n: not isinstance(n, Tree.TerminalNode))
 
+    def visitJoin_using_part(self, ctx):
+        return self.visitChildren(ctx, predicate = lambda n: not isinstance(n, Tree.TerminalNode))
+
+
     # converting case insensitive keywords to lowercase -----------------------
 
     def visitRegular_id(self, ctx):
@@ -311,11 +359,18 @@ class CustomErrorListener(ErrorListener):
         else:
             raise AntlrException(msg, None)
 
-    #def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
-    #    raise Exception("TODO")
+    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
+        return
+        #raise Exception("TODO")
 
-    #def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
-    #    raise Exception("TODO")
+    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
+        return
+        #raise Exception("TODO")
 
-    #def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
-    #    raise Exception("TODO")
+    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
+        return
+        #raise Exception("TODO")
+
+import pkg_resources
+speaker_cfg = yaml.load(pkg_resources.resource_stream('antlr_plsql', 'speaker.yml'))
+speaker = Speaker(**speaker_cfg)
